
package jworms;

import jworms.game.*;
import jworms.game.explosion.Explosion;
import jworms.game.explosion.FlameExplosion;
import jworms.game.weapons.*;
import jworms.openAL.OpenALListener;
import jworms.openAL.OpenALUtils;
import jworms.openAL.OpenALWave;
import jworms.openGL.BitmapFont;
import jworms.openGL.OpenGLTextureInternal;
import jworms.openGL.OpenGLUtils;
import jworms.network.UDPNetwork;
import jworms.network.UDPNetworkListener;
import jworms.network.GameInfo;
import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.GL;
import org.lwjgl.opengl.Window;

import java.net.SocketException;
import java.io.DataInputStream;
import java.io.IOException;

public class JWorms implements UDPNetworkListener{
  private BitmapFont font = null;

  // game to join
  private GameInfo gameInfo = null;

  // game field and objects
  private GameField gameField = null;

  private UDPNetwork network = null;
  private int networkID = 0;

  private Player player = null;
  private PlayerInformationUI playerInformationUI = null;
  private WeaponViewer weaponViewer = null;

  public static void main(String[] arguments){
    new JWorms(arguments.length > 0 && arguments[0].equalsIgnoreCase("host"));
  }

  public JWorms(boolean isHost){
    // network
    if(isHost){
      network = new UDPNetwork();
      network.addUDPNetworkListener(this);

      // host opens a game
      try{
        network.openGame("Worms LAN party!", 8);
      } catch(SocketException e){
        System.out.println("Error opening game : " + e.getMessage());
      }
    }else{
      network = new UDPNetwork();
      network.addUDPNetworkListener(this);

      // looks for games
      network.lookForGames();

      try{
        Thread.sleep(1000);
      } catch(InterruptedException e){ }

      if(gameInfo == null){
        System.out.println("Error: no games found.");
        System.exit(-1);
      }

      // join game
      System.out.println("Game: " + gameInfo);
      try{
        network.joinGame(gameInfo);
      } catch(IOException e){
        System.out.println("Error joining the game: " + e.getMessage());
      }
    }

    try{
      Thread.sleep(1000);
    } catch(InterruptedException e){ }

    if(networkID == 0 && !isHost){
      System.out.println("Error connecting to game.");
      System.exit(-1);
    }

    // init everything
    OpenGLUtils.initFullscreen("JWorms");
    OpenALUtils.alInit();

    GameFieldModifier.init();
    DesertEagle.init();
    AK47.init();
    RocketLauncher.init();
    Ball.init();
    Grenade.init();
    Explosion.init();
    FlameExplosion.init();
    Bullet.init();
    Rocket.init();

    Player.init();
    Worm.init();
    PlayerInformationUI.init();
    WeaponViewer.init();

    try{
      font = new BitmapFont("images/fonts/font.dat");
      gameField = new GameField("levels/LunarDig", network, networkID);
    } catch(Exception e){
      System.out.println("Failed starting game due to ");
      e.printStackTrace();
      System.exit(-1);
    }

    // generate all loaded textures
    OpenGLTextureInternal.glGenTextures();
    OpenALWave.alGenWaves();

    // create Player
    try{
      player = new Worm(false, "Worm", gameField);
      player.setKeyFocus(true);
      gameField.addGameObject(player);

      playerInformationUI = new PlayerInformationUI(player);

      weaponViewer = new WeaponViewer(player);
      weaponViewer.setWeapon(0, new DesertEagle(gameField));
      weaponViewer.setWeapon(1, new AK47(gameField));
      weaponViewer.setWeapon(2, new Grenade(gameField));
      weaponViewer.setWeapon(3, new RocketLauncher(gameField));
      weaponViewer.setWeapon(8, new Ball(gameField));
    } catch(Exception e){
      System.out.println("Failed starting game due to ");
      e.printStackTrace();
      System.exit(-1);
    }

    GameThread gameThread = new GameThread(gameField);
    gameThread.setDaemon(true);

    gameField.start(0);
    gameThread.start();

    while(!OpenGLUtils.isCloseRequested()){
      Mouse.poll();
      Keyboard.poll();

      gameThread.doMove();
      renderGameField();
      gameThread.waitUntilReady();
      renderObjects();

      // update internal window state
      Window.update();
      // swap buffers
      Window.paint();
    }

    gameThread.interrupt();

    // dispose Open GL and AL
    OpenGLUtils.destroy();
    OpenALUtils.alExit();
  }

  private void renderGameField(){
    // GL.glClear(GL.GL_COLOR_BUFFER_BIT);

    // GameField
    gameField.scrollTo(player.getPosition().x, player.getPosition().y);
    gameField.doScroll();

    GL.glDisable(GL.GL_BLEND);
    GL.glDisable(GL.GL_ALPHA_TEST);
    gameField.drawBack();
    GL.glEnable(GL.GL_ALPHA_TEST);
    gameField.drawFront();
    GL.glDisable(GL.GL_ALPHA_TEST);
    GL.glEnable(GL.GL_BLEND);
  }

  private void renderObjects(){
    // GameFieldObjects
    gameField.drawObjects();

    // PlayerInformationUI
    GL.glColor4f(1.0f, 1.0f, 1.0f, 0.5f);
    playerInformationUI.draw(font);
    GL.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

    // WeaponViewer
    weaponViewer.move();
    weaponViewer.draw();

    // update listener position
    OpenALListener.setPos(player.getPosition().x, player.getPosition().y, 100);
  }

// automatically join games
  public void gameJoined(int id){
    System.out.println("Game joined with id = " + id + " !");

    networkID = id;

    // we don't need to listen any more
    network.removeUDPNetworkListener(this);
  }

  public void receivePacket(DataInputStream in){
  }

  public void foundGame(GameInfo gameInfo){
    this.gameInfo = gameInfo;
  }
}
