
package jworms.game;

import jworms.game.GameField;
import jworms.game.GameObject;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.awt.*;
import java.awt.geom.Point2D;

public class GameFieldObjectFinder{
  /** Size of field tiles (horizontal and vertical). */
  private int tileSize = 64;
  /** Number of tiles horizontal. */
  private int tilesHorizontal = 0;
  /** Number of tiles vertical. */
  private int tilesVertical = 0;

  /** The game objects for every field tile. */
  private ArrayList tileObjects[][] = null;

  /** ArrayList to store search results (reuse is faster than recreation). */
  private ArrayList resultList = new ArrayList(100);

  /**
   * Creates a GameFieldObjectFinder for the given game field.
   *
   * @param gameField game field
   */
  public GameFieldObjectFinder(GameField gameField){
    // number of tiles horizontal/ vertical
    tilesHorizontal = (gameField.getWidth() + tileSize - 1) / tileSize;
    tilesVertical = (gameField.getHeight() + tileSize - 1) / tileSize;

    // initialize tiles
    tileObjects = new ArrayList[tilesHorizontal][tilesVertical];
    for(int tile_x = 0; tile_x < tilesHorizontal; tile_x++){
      for(int tile_y = 0; tile_y < tilesVertical; tile_y++){
        tileObjects[tile_x][tile_y] = new ArrayList();
      }
    }
  }

  /**
   * Returns the tile for the given object.
   *
   * @param position position
   * @return tile number (.x and .y)
   */
  private Point getTile(Point2D.Float position){
    int tileX = Math.max(0, Math.min((int) position.x / tileSize, tilesHorizontal - 1));
    int tileY = Math.max(0, Math.min((int) position.y / tileSize, tilesVertical - 1));

    return new Point(tileX, tileY);
  }

  /**
   * Adds a GameObject to the object finder.
   *
   * @param gameObject game object
   */
  public void add(GameObject gameObject){
    Point tile = getTile(gameObject.getPosition());
    tileObjects[tile.x][tile.y].add(gameObject);
  }

  /**
   * Add all GameObjects to the object finder.
   *
   * @param collection collection of game objects
   */
  public void addAll(Collection collection){
    Iterator iterator = collection.iterator();
    while(iterator.hasNext()){
      add((GameObject) iterator.next());
    }
  }

  /**
   * Removes a GameObject from the object finder.
   *
   * @param gameObject game object
   */
  public void remove(GameObject gameObject){
    Point tile = getTile(gameObject.getPosition());
    remove(gameObject, tile);
  }

  /**
   * Remove all GameObjects from the object finder.
   *
   * @param collection collection of game objects
   */
  public void removeAll(Collection collection){
    Iterator iterator = collection.iterator();
    while(iterator.hasNext()){
      remove((GameObject) iterator.next());
    }
  }

  /**
   * Removes the game object from the given tile.
   *
   * @param gameObject game object
   * @param tile tile
   */
  private void remove(GameObject gameObject, Point tile){
    if(tileObjects[tile.x][tile.y].contains(gameObject)){
      tileObjects[tile.x][tile.y].remove(gameObject);
    }else{
      System.out.println("Error in GameFieldObjectFinder.remove(): no such object.");
    }
  }

  /**
   * Called by the game field if an object has moved.
   *
   * @param gameObject
   */
  public void moved(GameObject gameObject){
    Point lastTile = getTile(gameObject.getLastPosition());
    Point tile = getTile(gameObject.getPosition());

    // nothing to do ?
    if(lastTile.equals(tile)) return;

    // remove at last position and add at new position
    remove(gameObject, lastTile);
    add(gameObject);
  }

  /**
   * Returns all objects within the given distance from position.
   *
   * @param position position
   * @param distance distance
   * @return list of objects
   */
  public ArrayList getGameObjects(Point2D.Float position, float distance){
    resultList.clear();

    Point upperLeft = getTile(new Point2D.Float(position.x - distance, position.y - distance));
    Point lowerRight = getTile(new Point2D.Float(position.x + distance, position.y + distance));

    for(int tile_x = upperLeft.x; tile_x < lowerRight.x; tile_x++){
      for(int tile_y = upperLeft.y; tile_y < lowerRight.y; tile_y++){
        Iterator iterator = tileObjects[tile_x][tile_y].iterator();
        while(iterator.hasNext()){
          GameObject gameObject = (GameObject) iterator.next();
          if(gameObject.getPosition().distance(position) > distance) continue;

          resultList.add(gameObject);
        }
      }
    }

    return resultList;
  }
}
