
package jworms.game;

import org.lwjgl.Display;

import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Iterator;

import jworms.util.MathUtil;
import jworms.game.util.CollosionUtil;

public class GameFieldUtil{
  /** BufferedImage to draw the explosion polygon */
  private static BufferedImage explosionImage = null;
  /** Graphics to draw to explosionImage */
  private static Graphics2D explosionImageGraphics = null;

  static{
    explosionImage = new BufferedImage(256, 256, BufferedImage.TYPE_BYTE_BINARY);
    explosionImageGraphics = explosionImage.createGraphics();
  }

  /**
   * Removes pixels from the dot map and foreground landscape.
   *
   * @param gameField game field
   * @param x x center of the explosion
   * @param y y center of the explosion
   * @param maxRadius radius of the explosion
   * @param maxStrength strength (0..1)
   */
  public static void doExplosion(GameField gameField, int x, int y, int maxRadius, float maxStrength){
    Polygon explosionPolygon = new Polygon();

    // number of steps for angle
    int angleSteps = Math.max(8, maxRadius / 2);

    // calculate depth for every angle
    float angle = 0;
    float angleIncrement = 2 * (float) Math.PI / angleSteps;
    for(int step_nr = 0; step_nr < angleSteps; step_nr++){
      float positionX = x;
      float positionY = y;
      float directionX = (float) Math.cos(angle);
      float directionY = (float) Math.sin(angle);

      // depth for this angle
      float strength = maxStrength;
      for(int radius = 0; radius < maxRadius && strength > 0; radius++){
        strength -= gameField.getHardness((int) positionX, (int) positionY);

        positionX += directionX;
        positionY += directionY;
      }

      // add point to explosionRing
      explosionPolygon.addPoint((int) (positionX - x), (int) (positionY - y));
      angle += angleIncrement;
    }

    gameField.addGameObject(new GameFieldModifier(false, gameField, x, y, maxRadius,  explosionPolygon));
  }

  /**
   * Removes pixels from the dot map and foreground landscape.
   *
   * @param gameField game field
   * @param x x center of the explosion
   * @param y y center of the explosion
   * @param maxRadius radius of the explosion
   * @param explosionPolygon polygon of the explosion shape
   */
  protected static void doExplosion(GameField gameField, int x, int y, int maxRadius, Polygon explosionPolygon){
    Point2D.Float explosionCenter = new Point2D.Float(x,y);

    // hit objects
    int hitRadius = maxRadius * 3;
    ArrayList gameObjects = gameField.getGameObjects(explosionCenter, hitRadius);

    Iterator gameObjectsIterator = gameObjects.iterator();
    while(gameObjectsIterator.hasNext()){
      GameObject gameObject = (GameObject) gameObjectsIterator.next();

      float distance = (float) explosionCenter.distance(gameObject.getPosition().x, gameObject.getPosition().y);
      Point2D.Float direction = new Point2D.Float((gameObject.getPosition().x - explosionCenter.x) / distance,
                                                  (gameObject.getPosition().y - explosionCenter.y) / distance);

      // calculate hit strength
      float strength = hitRadius / 10;
      float positionX = explosionCenter.x;
      float positionY = explosionCenter.y;
      for(int radius = 0; radius < distance && strength > 0; radius++){
        strength -= gameField.getHardness((int) positionX, (int) positionY);
        strength -= 0.1;

        positionX += direction.x;
        positionY += direction.y;
      }

      CollosionUtil.multiply(direction, 5);
      if(strength > 0) gameObject.hit(direction, Math.min(strength, 2));
    }

    // modify foreground image
    explosionPolygon.translate(x, y);
    Graphics2D graphics = gameField.getForegroundTexture().getGraphics();
    graphics.setComposite(AlphaComposite.Clear);
    graphics.fillPolygon(explosionPolygon);
    gameField.getForegroundTexture().update(x - maxRadius - 1, y - maxRadius - 1, 2 * maxRadius + 2, 2 * maxRadius + 2);
    explosionPolygon.translate(-x, -y);

    // modify dotmap
    int explosionImageWidth = explosionImage.getWidth();
    int explosionImageHeight = explosionImage.getHeight();

    explosionPolygon.translate(explosionImageWidth / 2, explosionImageHeight / 2);

    explosionImageGraphics.setBackground(Color.BLACK);
    explosionImageGraphics.clearRect(0, 0, explosionImageWidth, explosionImageHeight);
    explosionImageGraphics.setColor(Color.WHITE);
    explosionImageGraphics.fillPolygon(explosionPolygon);

    int fieldOffsetX = x - explosionImageWidth / 2;
    int fieldOffsetY = y - explosionImageHeight / 2;

    int left = explosionImageWidth / 2 - maxRadius - 1;
    int right = explosionImageWidth / 2 + maxRadius + 1;
    int top = explosionImageHeight / 2 - maxRadius - 1;
    int bottom = explosionImageHeight / 2 + maxRadius + 1;
    for(int imageY = top; imageY <= bottom; imageY++){
      for(int imageX = left; imageX <= right; imageX++){
        if(explosionImage.getRGB(imageX, imageY) == 0xFF000000) continue;

        gameField.setDot(imageX + fieldOffsetX, imageY + fieldOffsetY, (byte) 0);
      }
    }

    Point screen = gameField.fieldToScreen(x, y);
    gameField.shake(MathUtil.inBounds(0, maxRadius / 100.f - 0.1f, 1) * MathUtil.inBounds(0, 200 / (float) Math.sqrt(10 + Math.pow(screen.x - Display.getWidth()/2, 2) + Math.pow(screen.y - Display.getHeight()/2, 2)), 1));
  }
}
