
package jworms.game;

import jworms.game.util.CollosionEnvironment;
import jworms.game.util.CollosionUtil;
import jworms.game.weapons.Rocket;
import jworms.util.MathUtil;
import org.lwjgl.input.Keyboard;

import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * Worm as player.
 */
public class Worm extends Player{
// static
  /** ID for the object class. */
  private static int classID = 0;
  public int getClassID() { return classID; }
  public void setClassID(int id) { classID = id; }

  /**
   * Loads textures and sound.
   */
  public static void init(){
    WormUI.init();

    // register class id
    GameField.registerClass(new Worm(true, null, null));
  }

  /** State: IDLE. */
  public static final byte STATE_IDLE = 0;
  /** State: SNAKING. */
  public static final byte STATE_SNAKING = 1;
  /** State: JUMPING. */
  public static final byte STATE_JUMPUNG = 2;

// non static
  /** Worm UI. */
  private WormUI ui = null;

  /** State of the worm. */
  private byte state = STATE_IDLE;
  /** Looking direction of the worm. */
  private byte lookDirection = Player.DIRECTION_LEFT;

  /** Position of the worm (in the game field). */
  private Point2D.Float position = null;
  /** Movement speed/ direction vector. */
  private Point2D.Float motion;
  /** Maximum motion speed. */
  private int MAX_MOTION = 15;

  /** Worm leg collosion vector. */
  private Point2D.Float hitVector = new Point2D.Float();

  /** Frames until next burn. */
  private int framesUnitBurn = 0;

  /**
   * Creates a worm player.
   *
   * @param isHologram true if this is a hologram
   * @param name name of the worm
   * @param gameField game field
   */
  public Worm(boolean isHologram, String name, GameField gameField){
    super(isHologram, name, gameField);

    // set worm to the center of the game field
    position = new Point2D.Float(640, 50);
    motion = new Point2D.Float(0, 0);

    // create worm UI
    ui = new WormUI(this, gameField);
  }

  public Point2D.Float getPosition(){
    return position;
  }

  public void doAction(){
  }

  public void move(){
    super.move();

    // burning
    if(burnFactor > 0 && framesUnitBurn == 0){
      if(state != STATE_JUMPUNG){
        motion.y -= Math.random() * 3;
        motion.x += (Math.random() - 0.5) * 1;

        state = STATE_JUMPUNG;
      }

      health = Math.max(0, health - burnFactor / 50);
      framesUnitBurn = 20;
      burnFactor = 0;
    }
    if(framesUnitBurn > 0) framesUnitBurn--;


    CollosionEnvironment collosion = ui.getCollosion();

    Point walkZone = ui.getZone(WormUI.ZONE_WALK)[0];
    Point blockZone[] = ui.getZone(WormUI.ZONE_BLOCK);

    // go out of ground
    Point2D.Float walkPoint = new Point2D.Float(position.x + walkZone.x, position.y + walkZone.y);
    hitVector = CollosionUtil.getHitVector(walkPoint.x, walkPoint.y, collosion, gameField);
    float hitStrength = CollosionUtil.getLength(hitVector);

    // calculate hit point with ground
    Rectangle2D.Float hitPoint = CollosionUtil.getHitPoint(walkPoint, motion, collosion, gameField);
    if(hitPoint != null){
      hitPoint.x -= walkZone.x;
      hitPoint.y -= walkZone.y;
    }

    // closest hit
    for(int blockZone_nr = 0; blockZone_nr < blockZone.length; blockZone_nr++){
      Point zonePoint = blockZone[blockZone_nr];
      Point2D.Float blockPoint = new Point2D.Float(position.x + zonePoint.x, position.y + zonePoint.y);

      Rectangle2D.Float blockHitPoint = CollosionUtil.getHitPoint(blockPoint, motion, collosion, gameField);
      if(blockHitPoint == null) continue;

      blockHitPoint.x -= zonePoint.x;
      blockHitPoint.y -= zonePoint.y;

      if(hitPoint == null){
        hitPoint = blockHitPoint;
        continue;
      }

      float hitDistance = (float) Math.sqrt(Math.pow(position.x - hitPoint.x, 2) + Math.pow(position.y - hitPoint.y, 2));
      float blockDistance = (float) Math.sqrt(Math.pow(position.x - blockHitPoint.x, 2) + Math.pow(position.y - blockHitPoint.y, 2));
      if(blockDistance < hitDistance){
        hitPoint = blockHitPoint;
      }
    }

    if(hitPoint != null){
      // hit vector
      hitVector.x = hitPoint.width;
      hitVector.y = hitPoint.height;

      // stop shaking
      if(Math.abs(motion.x) < 0.3 && Math.abs(motion.y) < 0.5){
        motion.x = 0;
        motion.y = 0.001f - GameField.GRAVITY;

        state = STATE_IDLE;
      } else{
        // motion
        position.x = hitPoint.x;
        position.y = hitPoint.y;

        // hit on the ground
        Point2D.Float normal = CollosionUtil.getNormalized(hitVector);
        float hitNormalStrength = Math.abs(CollosionUtil.getScalarProduct(motion, normal));

        // motion manipulation
        motion.x += 2 * normal.x * hitNormalStrength;
        motion.y += 2 * normal.y * hitNormalStrength;

        // absobtion
        motion.x *= 0.2;
        motion.y *= 0.2;
      }
    } else{
      // hit vector
      hitVector.x = 0;
      hitVector.y = 0;

      // ball motion
      position.x += motion.x;
      position.y += motion.y;
    }

    // gravity
    motion.y += GameField.GRAVITY;

    // go out of ground handling (very tricky)
    if(hitStrength > 1 && (motion.x != 0 || Math.abs(motion.y - 0.001) > 0.001)){
      float dx = hitVector.x / (float) Math.sqrt(hitStrength);
      float dy = hitVector.y / (float) Math.sqrt(hitStrength);

      position.x += dx;
      position.y += dy;

      dx /= 10;
      dy /= 10;
      if((dx > 0 && motion.x < dx) || (dx < 0 && motion.x > dx)) motion.x += dx;
      if((dy > 0 && motion.y < dy) || (dy < 0 && motion.y > dy)) motion.y += dy;
    }

    // motion normalization
    motionNormalization();

    // motion slow down
    if(state != STATE_JUMPUNG || Math.abs(hitVector.y) > 0.5){
      motion.x *= 0.8;
    } else{
      motion.x *= 0.995;
    }
    motion.y *= 0.995;

    // motion interaction
    motionInteraction();
  }

  public void draw(){
    ui.draw();
  }

  public void dispose(){
    ui.dispose();
  }

  /**
   * User interaction with worm motion.
   */
  private void motionInteraction(){
    // jumping
    if(isKeyDown(KEY_MASK_JUMP) && (state != STATE_JUMPUNG || Keyboard.isKeyDown(Keyboard.KEY_LSHIFT))){
      motion.y = -4;

      motion.x = MathUtil.inBounds(-0.6f, getLongAverageMotion().x, 0.6f) * 2.5f;
      if(isKeyDown(KEY_MASK_MOVE_LEFT)){
        if(motion.x > 0) motion.x = 0;
        motion.x -= 1;
      }
      if(isKeyDown(KEY_MASK_MOVE_RIGHT)){
        if(motion.x < 0) motion.x = 0;
        motion.x += 1;
      }

      state = STATE_JUMPUNG;
      return;
    }

    if(state != STATE_JUMPUNG && Math.abs(hitVector.y) < 0.2 && Math.abs(getShortAverageMotion().y) > 1){
      state = STATE_JUMPUNG;
    }
    if(state == STATE_JUMPUNG && Math.abs(hitVector.y) > 0.5 && Math.abs(getShortAverageMotion().y) < 1){
      state = STATE_SNAKING;
    }

    // worm is skidding
    if(Math.abs(motion.x) > 1.1) return;

    // move left or right
    if(state != STATE_JUMPUNG){
      if(isKeyDown(KEY_MASK_MOVE_LEFT)){
        motion.x = -0.3f / Math.max(0.3f, -getShortAverageMotion().y);
        lookDirection = DIRECTION_LEFT;

        state = STATE_SNAKING;
      }
      if(isKeyDown(KEY_MASK_MOVE_RIGHT)){
        motion.x = 0.3f / Math.max(0.3f, -getShortAverageMotion().y);
        lookDirection = DIRECTION_RIGHT;

        state = STATE_SNAKING;
      }
    }
  }

  /**
   * Keep motion in normal bounds.
   */
  private void motionNormalization(){
    float speed = (float) (Math.sqrt(Math.pow(motion.x, 2) + Math.pow(motion.y, 2)));

    if(speed > MAX_MOTION){
      motion.x *= MAX_MOTION / speed;
      motion.y *= MAX_MOTION / speed;
    }
  }

  /**
   * Returns the motion vector of the worm.
   *
   * @return motion vector
   */
  public Point2D.Float getMotion(){
    return motion;
  }

  /**
   * Return the direction of the worms motion.
   *
   * @return motion direction
   */
  public int getDirection(){
    if(motion.x < 0) return DIRECTION_LEFT;
    if(motion.x > 0) return DIRECTION_RIGHT;

    return lookDirection;
  }

  /**
   * Returns the looking direction of the worm.
   *
   * @return looking direction
   */
  public int getLookDirection(){
    return lookDirection;
  }

  /**
   * Returns the state of the worm.
   *
   * @return state
   */
  public byte getState(){
    return state;
  }

// object interaction
  public void hit(Point2D.Float motion, float weight){
    this.motion.x += motion.x * weight;
    this.motion.y += motion.y * weight;

    health = Math.max(0, health - Math.min(50, CollosionUtil.getLength(motion) * weight / 50));
  }

// network
  public GameObject create(GameField gameField, DataInputStream in) throws IOException{
    Worm worm = new Worm(true, "", gameField);
    worm.recieveUpdate(in);
    return worm;
  }

  public void sendUpdate(DataOutputStream out) throws IOException{
    super.sendUpdate(out);

    out.writeFloat(position.x);
    out.writeFloat(position.y);
    out.writeFloat(motion.x);
    out.writeFloat(motion.y);
    out.writeByte(state);
    out.writeByte(lookDirection);
  }

  public void recieveUpdate(DataInputStream in) throws IOException{
    super.recieveUpdate(in);

    position.x = in.readFloat();
    position.y = in.readFloat();
    motion.x = in.readFloat();
    motion.y = in.readFloat();
    state = in.readByte();
    lookDirection = in.readByte();
  }
}
