
package jworms.game;

import jworms.game.util.CollosionEnvironment;
import jworms.game.util.CollosionUtil;
import jworms.openGL.SmallTexture;
import org.lwjgl.opengl.GL;

import java.awt.*;
import java.awt.image.PixelGrabber;
import java.util.ArrayList;

public class WormUI extends PlayerUI{
// static
  /** Number of zones. */
  public static final int ZONES_COUNT = 2;

  /** Zone for walking (bottom of the legs). */
  public static final int ZONE_WALK = 0;
  /** Zone for blocking (front of the body). */
  public static final int ZONE_BLOCK = 1;

  /** Textures for the snaking worm. */
  private static SmallTexture textures_snaking[] = null;

  /**
   * Loads textures and sound.
   */
  public static void init(){
    // load images
    try{
      // load snaking textures
      textures_snaking = loadTextures("images/worm/worm_sneaking");
    } catch(Exception e){
      System.out.println("Error in BallUI during loading images.");
      e.printStackTrace();
    }
  }

  /**
   * Loads textures with the given prefix name.
   * (textureName_1.png, textureName_2.png, ...)
   *
   * @param textureName prefix name
   * @return textures array
   */
  public static SmallTexture[] loadTextures(String textureName){
    ArrayList texturesList = new ArrayList();

    int textureNr = 1;
    while(true){
      try{
        SmallTexture texture = new SmallTexture(textureName + "_" + textureNr + ".png", GL.GL_LINEAR);
        texturesList.add(texture);
      } catch(Exception e){
        break;
      }
      textureNr++;
    }

    SmallTexture textures[] = new SmallTexture[texturesList.size()];
    texturesList.toArray(textures);

    return textures;
  }

// non static
  /** The associated worm. */
  private Worm worm = null;
  /** Game field. */
  private GameField gameField = null;

  /** The zones of the worm: Point[] zone[direction {left = 0, right = 1}][zone {ZONE_...}]. */
  private Point zones[][][] = new Point[2][2][];
  /** Collosion points. */
  private CollosionEnvironment collosion = null;

  /**
   * Creates a UI for the given worm.
   *
   * @param worm worm
   * @param gameField game field
   */
  public WormUI(Worm worm, GameField gameField){
    super(worm, gameField);

    this.worm = worm;
    this.gameField = gameField;

    // load worm zones
    loadZones("images/worm/worm_zones.png");

    // calculate zone neighborhood
    collosion = CollosionUtil.createNeighborhood(9, 11);
  }

  /**
   * Loads the worm zones from image.
   *
   * @param imageName image name
   */
  private void loadZones(String imageName){
    Image zoneImage = (new javax.swing.ImageIcon(imageName)).getImage();
    int width = zoneImage.getWidth(null);
    int height = zoneImage.getHeight(null);

    PixelGrabber grabber = new PixelGrabber(zoneImage, 0, 0, width, height, false);
    try{
      grabber.grabPixels();
      byte pixelArray[] = (byte[]) grabber.getPixels();

      for(int zone_nr = 0; zone_nr < ZONES_COUNT; zone_nr++){
        // collect the points that belong to the zone
        ArrayList zonePoints = new ArrayList();
        for(int y = 0; y < height; y++){
          for(int x = 0; x < width; x++){
            if(pixelArray[x + y * width] != zone_nr + 1) continue;

            zonePoints.add(new Point(x, y));
          }
        }

        // convert into array for both directions
        zones[0][zone_nr] = new Point[zonePoints.size()];
        zones[1][zone_nr] = new Point[zonePoints.size()];
        for(int point_nr = 0; point_nr < zonePoints.size(); point_nr++){
          Point point = (Point) zonePoints.get(point_nr);

          zones[0][zone_nr][point_nr] = new Point(point.x - width / 2, point.y - height / 2);
          zones[1][zone_nr][point_nr] = new Point(width / 2 - point.x, point.y - height / 2);
        }
      }
    } catch(Exception e){
      System.out.println("Loading zone image failed due to " + e);
      System.exit(-1);
    }
  }

  /**
   * Returns the zone point for the given zone.
   *
   * @param zone_nr zone (ZONE_...)
   * @return zone points
   */
  public Point[] getZone(int zone_nr){
    return zones[worm.getDirection()][zone_nr];
  }

  /**
   * Returns the collosion environment (points).
   *
   * @return collosion points
   */
  public CollosionEnvironment getCollosion(){
    return collosion;
  }

  /**
   * Draws the worm.
   */
  public void draw(){
    Point position = new Point((int) worm.getInterpolatedPosition().x, (int) worm.getInterpolatedPosition().y);

    int texture_nr = 0;
    if(worm.getState() == Worm.STATE_SNAKING){
      texture_nr = (int) ((System.currentTimeMillis() / 120) % textures_snaking.length);
    }

    SmallTexture texture = textures_snaking[texture_nr];
    int width = texture.getImage().getWidth();
    int height = texture.getImage().getHeight();

    // get screen position and draw the texture
    Point screen = gameField.fieldToScreen(position.x - width / 2, position.y - height / 2);
    texture.draw(screen.x, screen.y, worm.getLookDirection() == Player.DIRECTION_LEFT ? -1.0f : 1.0f, 1.0f, 0.0f);

    super.draw();
  }

  /**
   * Free all resources associated with this object.
   */
  public void dispose(){
    super.dispose();
  }
}
