
package jworms.game.explosion;

import jworms.game.GameField;
import jworms.game.GameObject;
import jworms.util.MathUtil;
import org.lwjgl.opengl.GL;

import java.awt.geom.Point2D;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

public class FlameExplosion extends GameObject{
// static
  /** ID for the object class. */
  private static int classID = 0;
  public int getClassID() { return classID; }
  public void setClassID(int id) { classID = id; }

  /**
   * Loads textures and sound.
   */
  public static void init(){
    FlameExplosionFlame.init();
    FlameExplosionGlow.init();

    // register class id
    GameField.registerClass(new FlameExplosion());
  }

  /** ArrayList of all explosion objects from all flame explosions. */
  private static ArrayList allExplosionObjects = new ArrayList();
  /** Current explosion object index for heat calculation. */
  private static int heatCalculationIndex = 0;

  /**
   * Adds an explosion object to the list of all explosion objects.
   *
   * @param explosionObject explosion object
   */
  private static synchronized void addExplosionObject(ExplosionObject explosionObject){
    allExplosionObjects.add(explosionObject);
  }

  /**
   * Removes an explosion object from the list of all explosion objects.
   *
   * @param explosionObject explosion object
   */
  private static synchronized void removeExplosionObject(ExplosionObject explosionObject){
    allExplosionObjects.remove(explosionObject);
  }

  /**
   * Calculates the heat for exlosion objects.
   */
  private static synchronized void heatCalculation(){
    if(allExplosionObjects.isEmpty()) return;

    for(int calculation_nr = 0; calculation_nr < 5; calculation_nr++){
      if(heatCalculationIndex >= allExplosionObjects.size()) heatCalculationIndex = 0;
      ExplosionObject calculationObject = (ExplosionObject) allExplosionObjects.get(heatCalculationIndex);

      // calculate the heat of the object
      calculationObject.heat = 0;
      for(int object_nr = 0; object_nr < allExplosionObjects.size(); object_nr++){
        ExplosionObject explosionObject = (ExplosionObject) allExplosionObjects.get(object_nr);

        float distance = Math.abs(calculationObject.position.x - explosionObject.position.x) +
                Math.abs(calculationObject.position.y - explosionObject.position.y);

        if(distance < 10) calculationObject.heat += distance / 100;
      }

      heatCalculationIndex++;
    }
  }

// non static
  /** GameField */
  private GameField gameField = null;
  /** Position of the explosion. */
  private Point2D.Float position = null;

  /** Time when the exlosion started. */
  private long startTime = 0;
  /** Seconds of livetime. */
  protected float lifeSeconds = 0;

  /** ExplosionObjects */
  private ArrayList explosionObjects = new ArrayList();

  /**
   * Creates an explosion.
   *
   * @param isHologram true if this is a hologram
   * @param gameField game field
   * @param position position
   */
  public FlameExplosion(boolean isHologram, GameField gameField, Point2D.Float position){
    super(isHologram);

    this.gameField = gameField;
    this.position = position;

    startTime = System.currentTimeMillis();

    // number of steps for angle
    int angleSteps = 196;

    // calculate depth for every angle
    float angle = 0;
    float angleIncrement = 2* (float) Math.PI / angleSteps;

    for(int step_nr = 0; step_nr < angleSteps; step_nr++){
      float flameSpeed = 0.5f + (float) Math.random()*10;

      Point2D.Float flamePosition = new Point2D.Float(position.x, position.y);
      Point2D.Float flameMotion = new Point2D.Float(flameSpeed * (float) Math.cos(angle), flameSpeed * ((float) Math.sin(angle) - 0.9f));

      ExplosionObject flame = null;
      if(step_nr % 2 == 0){
        flame = new FlameExplosionFlame(gameField, flamePosition, flameMotion);
      }else{
        flame = new FlameExplosionGlow(gameField, flamePosition, flameMotion);
      }
      explosionObjects.add(flame);
      addExplosionObject(flame);

      angle += angleIncrement;
    }
  }

  /**
   * Creates a explosion that will not be usable as game object.
   */
  public FlameExplosion(){
    super(true);
  }

  public Point2D.Float getPosition(){
    return position;
  }

  public void doAction(){
    gameField.removeGameObject(this);
  }

  public void doMove(){
    lifeSeconds = (System.currentTimeMillis() - startTime) / 1000.f;

    // calculate the heat for the flame
    heatCalculation();

    Iterator explosionObjectsIterator = explosionObjects.iterator();
    while(explosionObjectsIterator.hasNext()){
      ExplosionObject explosionObject = (ExplosionObject) explosionObjectsIterator.next();
      explosionObject.doMove();

      // burn other object
      if(explosionObject.lifeSeconds < explosionObject.maxLife - 1){
        ArrayList gameObjects = gameField.getGameObjects(explosionObject.position, 40);

        Iterator gameObjectsIterator = gameObjects.iterator();
        while(gameObjectsIterator.hasNext()){
          GameObject gameObject = (GameObject) gameObjectsIterator.next();

          double distanceFoot = explosionObject.position.distance(gameObject.getPosition().x, gameObject.getPosition().y + 15);
          double distanceHead = explosionObject.position.distance(gameObject.getPosition().x, gameObject.getPosition().y + -5);
          if(distanceFoot > 15 && distanceHead > 15) continue;

          gameObject.burn(MathUtil.inBounds(0.2f, explosionObject.heat, 1));
        }
      }

      // let the flame die if the heat is very low
      if(explosionObject.heat < 0.2) explosionObject.maxLife -= 0.03;

      if(explosionObject.isFinished()){
        explosionObjectsIterator.remove();
        removeExplosionObject(explosionObject);
      }
    }

    // dispose explosion
    if(!isHologram && lifeSeconds > 20) hasAction = true;
  }

  public void draw(){
    GL.glDisable(GL.GL_ALPHA_TEST);
    GL.glEnable(GL.GL_BLEND);

    // draw
    Iterator explosionObjectsIterator = explosionObjects.iterator();
    while(explosionObjectsIterator.hasNext()){
      ExplosionObject explosionObject = (ExplosionObject) explosionObjectsIterator.next();
      explosionObject.draw();
    }

    // draw glow
    Iterator glowIterator = explosionObjects.iterator();
    while(glowIterator.hasNext()){
      ExplosionObject explosionObject = (ExplosionObject) glowIterator.next();
      if(!(explosionObject instanceof FlameExplosionGlow)) continue;

      ((FlameExplosionGlow) explosionObject).drawGlow();
    }

    GL.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
  }

  public void dispose(){
  }

// network
  public GameObject create(GameField gameField, DataInputStream in) throws IOException{
    FlameExplosion flameExplosion = new FlameExplosion(true, gameField, new Point2D.Float());
    flameExplosion.recieveUpdate(in);
    return flameExplosion;
  }

  public void sendUpdate(DataOutputStream out) throws IOException{
    out.writeFloat(position.x);
    out.writeFloat(position.y);
  }

  public void recieveUpdate(DataInputStream in) throws IOException{
    position.x = in.readFloat();
    position.y = in.readFloat();
  }
}
