
package jworms.game.util;

import jworms.game.GameField;

import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

public class CollosionUtil{
  /**
   * Creates a ring of points (neighborhood) between min and max radius.
   *
   * @param minRadius minimum radius
   * @param maxRadis maximum radius
   * @return Ring of points that lie inside minRadius <= raduis <= maxRadius
   */
  public static CollosionEnvironment createNeighborhood(float minRadius, float maxRadis){
    int radius = (int) (maxRadis + 1);

    // caculate neighborhood points
    ArrayList neighborsList = new ArrayList();
    for(int y = -radius; y <= radius; y++){
      for(int x = -radius; x <= radius; x++){
        float distance = (float) Math.sqrt(Math.pow(x, 2) + Math.pow(y, 2));
        if(distance < minRadius || distance > maxRadis) continue;

        neighborsList.add(new Point(x, y));
      }
    }

    CollosionEnvironment collosion = new CollosionEnvironment();
    collosion.neighborhood = new Point[neighborsList.size()];
    neighborsList.toArray(collosion.neighborhood);

    // calculate normalization
    Point2D.Float normalizationVector = new Point2D.Float();

    for(int point_nr = 0; point_nr < collosion.neighborhood.length; point_nr++){
      Point point = collosion.neighborhood[point_nr];

      if(point.y < maxRadis - 3) continue;

      if(point.x < 0) normalizationVector.x += -1. / (-2. + point.x);
      if(point.x > 0) normalizationVector.x += -1. / (2. + point.x);
      if(point.y < 0) normalizationVector.y += -1. / (-2. + point.y);
      if(point.y > 0) normalizationVector.y += -1. / (2. + point.y);
    }

    collosion.normalizationFactor = maxRadis > 5 ? 1 / getLength(normalizationVector) : 2;

    return collosion;
  }

  /**
   * Calculates the hit vector with the given neighborhood.
   *
   * @param x x
   * @param y y
   * @param collosion collosion environmen
   * @param gameField game field
   * @return hit vector
   */
  public static Point2D.Float getHitVector(float x, float y, CollosionEnvironment collosion, GameField gameField){
    Point2D.Float hitVector = new Point2D.Float();

    // calcualte hit vector
    for(int point_nr = 0; point_nr < collosion.neighborhood.length; point_nr++){
      Point point = collosion.neighborhood[point_nr];

      if(gameField.getDot((int) x + point.x, (int) y + point.y) == GameField.DOTMAP_BACKGROUND) continue;

      if(point.x < 0) hitVector.x += -1. / (-2. + point.x);
      if(point.x > 0) hitVector.x += -1. / (2. + point.x);
      if(point.y < 0) hitVector.y += -1. / (-2. + point.y);
      if(point.y > 0) hitVector.y += -1. / (2. + point.y);
    }

    // normalize hit vector
    multiply(hitVector, collosion.normalizationFactor);

    return hitVector;
  }

  /**
   * Calculates the absolute hit vector with the given neighborhood.
   *
   * @param x x
   * @param y y
   * @param collosion collosion environmen
   * @param gameField game field
   * @return hit vector
   */
  public static Point2D.Float getAbsoluteHitVector(float x, float y, CollosionEnvironment collosion, GameField gameField){
    Point2D.Float absoluteHitVector = new Point2D.Float();

    // calcualte hit vector
    for(int point_nr = 0; point_nr < collosion.neighborhood.length; point_nr++){
      Point point = collosion.neighborhood[point_nr];

      if(gameField.getDot((int) x + point.x, (int) y + point.y) == GameField.DOTMAP_BACKGROUND) continue;

      if(point.x < 0) absoluteHitVector.x += -1. / (-2. + point.x);
      if(point.x > 0) absoluteHitVector.x += 1. / (2. + point.x);
      if(point.y < 0) absoluteHitVector.y += -1. / (-2. + point.y);
      if(point.y > 0) absoluteHitVector.y += 1. / (2. + point.y);
    }

    // normalize hit vector
    multiply(absoluteHitVector, collosion.normalizationFactor);

    return absoluteHitVector;
  }

  /**
   * Calculates the hit point with the ground.
   * Returns null if the object does not hit the ground.
   *
   * @param point position of the object
   * @param motion motion of the point (x, y)
   * @param collosion collosion environment
   * @param gameField game field
   *
   * @return hit point with the ground (rect.x, rect.y) and the hit vector in this point (rect.width, rect.height)
   */
  public static Rectangle2D.Float getHitPoint(Point2D.Float point, Point2D.Float motion, CollosionEnvironment collosion, GameField gameField){
    // strength of hit on the ground
    float hitStrength = 1.5f;

    // hitVector at start position
    float lastStrength = getLength(getHitVector(point.x, point.y, collosion, gameField));

    // current position and distance to move
    Point2D.Float position = new Point2D.Float(point.x, point.y);
    float distance = getLength(motion);

    // move only distance 1 per step
    Point2D.Float stepMotion = distance != 0 ?
            new Point2D.Float(motion.x / distance, motion.y / distance) : new Point2D.Float();
    Point2D.Float hitVector = null;
    for(int step_nr = 0; step_nr < distance + 1; step_nr++){
      hitVector = getHitVector(position.x, position.y, collosion, gameField);
      ;
      float strength = getLength(hitVector);

      // did we hit on the ground ?
      if(strength > hitStrength && strength > lastStrength && getScalarProduct(hitVector, stepMotion) <= 0){
        if(step_nr == 1 && lastStrength > hitStrength){
          position.x -= stepMotion.x;
          position.y -= stepMotion.y;
        }
        return new Rectangle2D.Float(position.x, position.y, hitVector.x, hitVector.y);
      }

      // move distance 1
      position.x += stepMotion.x;
      position.y += stepMotion.y;

      lastStrength = strength;
    }

    // we did not hit on the ground
    return null;
  }

  /**
   * Returns the length of the given vector.
   *
   * @param vector vector
   * @return length of the vector
   */
  public static float getLength(Point2D.Float vector){
    return (float) vector.distance(0, 0);
  }

  /**
   * Returns the normalized vector (length 1).
   *
   * @param vector vector
   * @return vector with same direction but length 1
   */
  public static Point2D.Float getNormalized(Point2D.Float vector){
    float length = getLength(vector);
    return length == 0 ? new Point2D.Float() : new Point2D.Float(vector.x / length, vector.y / length);
  }

  /**
   * Returns the scalar product of the two vectors.
   *
   * @param vector1 first vector
   * @param vector2 second vector
   * @return scalar product
   */
  public static float getScalarProduct(Point2D.Float vector1, Point2D.Float vector2){
    return vector1.x * vector2.x + vector1.y * vector2.y;
  }

  /**
   * Returns the cross product of the two vectors.
   *
   * @param vector1 first vector
   * @param vector2 second vector
   * @return cross product
   */
  public static float getCrossProduct(Point2D.Float vector1, Point2D.Float vector2){
    return vector1.x * vector2.y - vector1.y * vector2.x;
  }

  /**
   * Multiplies the given vector with factor.
   *
   * @param vector vector
   * @param factor factor
   */
  public static void multiply(Point2D.Float vector, float factor){
    vector.x *= factor;
    vector.y *= factor;
  }
}
