
package jworms.game.weapons;

import jworms.game.GameField;
import jworms.game.Player;
import jworms.game.util.CollosionEnvironment;
import jworms.game.util.CollosionUtil;
import jworms.openAL.OpenALSource;
import jworms.openAL.OpenALWave;
import jworms.openGL.SmallTexture;
import jworms.util.MathUtil;
import org.lwjgl.openal.AL;
import org.lwjgl.opengl.GL;

import java.awt.*;

public class BallUI{
// static
  /** Texture. */
  private static SmallTexture texture = null;

  /** Hit sound. */
  private static OpenALWave soundHit = null;

  /**
   * Loads textures and sound.
   */
  public static void init(){
    // load images
    try{
      texture = new SmallTexture("images/weapons/ball/ball.png", GL.GL_LINEAR);
    } catch(Exception e){
      System.out.println("Error in BallUI during loading images.");
      e.printStackTrace();
    }

    // load sound
    try{
      soundHit = null;
    } catch(Exception e){
      System.out.println("Error in BallUI during loading sound.");
      e.printStackTrace();
    }
  }

// non static
  /** The associated ball. */
  private Ball ball = null;
  /** Game field. */
  private GameField gameField = null;

  /** Collosion points. */
  private CollosionEnvironment collosion = null;

  /** Sound source. */
  protected OpenALSource soundSource = new OpenALSource();

  /** Alpha for drawing the ball. */
  protected float alpha = 0.5f;

  /**
   * Creates a UI for the given ball.
   *
   * @param ball ball
   * @param gameField game field
   */
  public BallUI(Ball ball, GameField gameField){
    this(ball, gameField, 15, 18);
  }

  /**
   * Creates a UI for the given ball.
   *
   * @param ball ball
   * @param gameField game field
   */
  protected BallUI(Ball ball, GameField gameField, float collosionMinRadius, float collosionMaxRadius){
    this.ball = ball;
    this.gameField = gameField;

    // calculate zone neighborhood
    collosion = CollosionUtil.createNeighborhood(collosionMinRadius, collosionMaxRadius);
  }

  /**
   * Returns the texture.
   *
   * @return texture
   */
  public SmallTexture getTexture(){
    return texture;
  }

  /**
   * Return the hit sound.
   *
   * @return hit sound
   */
  public OpenALWave getSoundHit(){
    return soundHit;
  }

  /**
   * Returns the collosion environment (points).
   *
   * @return collosion points
   */
  public CollosionEnvironment getCollosion(){
    return collosion;
  }

  /**
   * Draws the ball.
   */
  public void draw(){
    draw((int) ball.getInterpolatedPosition().x, (int) ball.getInterpolatedPosition().y, 1, 1, ball.getAngle());
  }

  /**
   * Draws a symbol for the weapon.
   *
   * @param player player
   */
  public void drawSymbol(Player player){
    int x = (int) player.getInterpolatedPosition().x + 0 * (player.getLookDirection() == Player.DIRECTION_RIGHT ? 1 : -1);
    int y = (int) player.getInterpolatedPosition().y + 8;

    draw(x, y, 0.5f, 0.5f, 0);
  }

  /**
   * Draws the texture at the given position.
   *
   * @param x x
   * @param y y
   * @param scaleHorizontal scale horizontal
   * @param scaleVertical scale vertical
   * @param angle angle (in radian)
   */
  private void draw(int x, int y, float scaleHorizontal, float scaleVertical, float angle){
    GL.glColor4f(1.0f, 1.0f, 1.0f, alpha);

    int width = getTexture().getImage().getWidth();
    int height = getTexture().getImage().getHeight();

    Point screen = gameField.fieldToScreen(x - width / 2, y - height / 2);
    getTexture().draw(screen.x, screen.y, scaleHorizontal, scaleVertical, (float) (angle * 180 / Math.PI));

    GL.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
  }

  /**
   * Free all resources associated with this object.
   */
  public void dispose(){
    soundSource.dispose();
  }

  /**
   * Called, when the ball hits the ground.
   */
  public void hitOnGround(float hitStrength){
    // play sound
    if(getSoundHit() != null && soundSource.bind()){
      if(soundSource.getPropi(AL.AL_SOURCE_STATE) == AL.AL_PLAYING) soundSource.stop();

      soundSource.setPos(ball.getPosition().x, ball.getPosition().y, 0);
      soundSource.setPropf(AL.AL_GAIN, MathUtil.inBounds(0, hitStrength / 6, 1));
      soundSource.play(getSoundHit());
    }
  }
}
