
package jworms.openAL;

import org.lwjgl.openal.AL;

import java.nio.IntBuffer;

public class OpenALSource{
// static
  /** Maximal count of sources. */
  protected static final int MAX_SOURCES = 64;

  /** IntBuffer of all available sources. */
  private static IntBuffer sources;
  /** Array of the appendant OpenALSources. */
  private static OpenALSource[] openALSources;


  /**
   * Set the reference distance, the rolloff factor and the maximal distance of the source.
   *
   * @param id the id of the source
   */
  private static void setSoundEnvironment(int id){
    AL.alSourcef(id, AL.AL_REFERENCE_DISTANCE, 100.0f);
    AL.alSourcef(id, AL.AL_ROLLOFF_FACTOR, 0.5f);
    AL.alSourcef(id, AL.AL_MAX_DISTANCE, 1000);
  }

  /**
   * Resets properties of the source.
   *
   * @param id the id of the source
   */
  public static void resetSource(int id){
    AL.alSourcef(id, AL.AL_GAIN, 1);
  }

  /**
   * Allocate all available sources.
   */
  public static void alGenSoureces(){
    // get all available sources
    for(int i = MAX_SOURCES; i > 0; i--){
      sources = OpenALUtils.createIntBuffer(i);
      AL.alGenSources(sources);
      if(AL.alGetError() == AL.AL_NO_ERROR) break;
    }

    // create empty openALSources
    openALSources = new OpenALSource[sources.capacity()];
    for(int i = 0; i < openALSources.length; i++){
      openALSources[i] = new OpenALSource();
      setSoundEnvironment(sources.get(i));
      openALSources[i].winControl(sources.get(i));
    }

//    System.out.println("Jeah! I get " + sources.capacity() + " sources.");
  }

  /**
   * Frees all allocated sources.
   */
  public static void alDeleteSources(){
    // free all allocated sources
    AL.alDeleteSources(sources);
  }

// non static
  /** The source id (name). */
  private int id;
  /** Do I have the control over the source? Can I play some waves now? */
  private boolean hasControl = false;

  /**
   * Creates an openAL source.
   */
  public OpenALSource(){
  }

  /**
   * Return the id of the source.
   *
   * @return id of the source
   */
  public int getId(){
    return id;
  }

  /**
   * Try to get control over an unused source.
   *
   * @return true if we have now control over an source
   */
  public boolean bind(){
    if(hasControl) return true;

    // search for non playing source
    for(int i = 0; i < openALSources.length; i++){
      OpenALSource openALSource = openALSources[i];
      int state = openALSource.getPropi(AL.AL_SOURCE_STATE);
      if(state != AL.AL_PLAYING && state != AL.AL_PAUSED){
        // we've found an unused source
        int id = openALSource.getId();
        openALSource.loseControl();
        resetSource(id);
        this.winControl(id);
        this.setPropf(AL.AL_GAIN, 1);
//        System.out.println("use sound source " + i + " for source id " + id);
//        System.out.println(i + ": " + this);
        return true;
      }
    }

    System.out.println("no sound source available");
    return false;
  }

  /**
   * The openALSource get control over an source.
   *
   * @param id the new id (name)
   */
  private void winControl(int id){
    this.id = id;
    hasControl = true;
  }

  /**
   * The openALSource lose it's control over the source.
   */
  protected void loseControl(){
    hasControl = false;
  }

  /**
   * Returns true if the openALSource has the control over a source.
   *
   * @return true if the openALSource has the control over a source
   */
  public boolean hasControl(){
    return hasControl;
  }

  public int getPropi(int prop){
    return AL.alGetSourcei(id, prop);
  }

  public float getPropf(int prop){
    return AL.alGetSourcef(id, prop);
  }

  public void setPropi(int prop, int value){
    AL.alSourcei(id, prop, value);
  }

  public void setPropf(int prop, float value){
    AL.alSourcef(id, prop, value);
  }

  public void setProp3f(int prop, float value1, float value2, float value3){
    AL.alSource3f(id, prop, value1, value2, value3);
  }

  public void setPos(float x, float y, float z){
    AL.alSource3f(id, AL.AL_POSITION, x, y, z);
  }

  /**
   * Start playing the wave file.
   *
   * @param openALWave the wave file.
   */
  public void play(OpenALWave openALWave){
    // set up source input
    AL.alSourcei(id, AL.AL_BUFFER, openALWave.id);
    OpenALUtils.checkError();

    // play source 0
    AL.alSourcePlay(id);
//    System.out.println("source id " + id + " is playing sound " + openALWave);
    OpenALUtils.checkError();
  }

  public void pause(){
    AL.alSourcePause(id);
  }

  public void stop(){
    AL.alSourceStop(id);
  }

  public void rewind(){
    AL.alSourceRewind(id);
  }

  public void dispose(){
  }

  public String toString(){
    return "ID: " + id + ", HasControl: " + (hasControl ? "yes" : "no");
  }
}
