
package jworms.openGL;

import org.lwjgl.opengl.GL;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;

public class BitmapFont{
  /** SmallTextures with font data. */
  private ArrayList fontTextures = null;
  /** Bitmap characters. */
  private BitmapFontCharacter characters[] = null;

  /** Top line of the font. */
  private int fontTop = 0;
  /** Bottom line of the font. */
  private int fontBottom = 0;
  /** Width of a space. */
  private int spaceWidth = 10;

  /**
   * Creates a BitmapFont.
   */
  public BitmapFont(String path) throws Exception{
    fontTextures = new ArrayList();
    characters = new BitmapFontCharacter[256];

    try{
      File file = new File(path);
      BufferedReader reader = new BufferedReader(new FileReader(file));

      path = file.getCanonicalPath();
      path = path.substring(0, path.lastIndexOf(File.separatorChar) + 1);

      String bimapFileName = null;
      while((bimapFileName = reader.readLine()) != null){
        int lineHeight = new Integer(reader.readLine()).intValue();

        String characterNames = reader.readLine();
        if(characterNames == null) break;

        SmallTexture texture = new SmallTexture(path + bimapFileName, GL.GL_LINEAR);
        fontTextures.add(texture);

        // extract letters
        extractLetters(texture, lineHeight, characterNames);
      }
    } catch(Exception e){
      System.out.println("Failed to read font due to " + e);
      throw e;
    }
  }

  /**
   * Extract the letters from the texture.
   *
   * @param texture texture
   * @param lineHeight line height in the texture
   * @param characterNames characters in the texture
   */
  private void extractLetters(SmallTexture texture, int lineHeight, String characterNames){
    // search for the first letter
    Point searchPosition = new Point(0, lineHeight / 2);

    int letter_nr = 0;
    Point letterPosition = null;
    while((letterPosition = getNextLetterPosition(texture, searchPosition, lineHeight)) != null){
      Rectangle letterRectangle = getLetterRectangle(texture, letterPosition);

      BitmapFontCharacter character = new BitmapFontCharacter();
      character.characterName = characterNames.charAt(letter_nr);
      character.texture = new SmallTexturePart(texture, letterRectangle.x, letterRectangle.y,
                                               letterRectangle.width, letterRectangle.height);

      // shrink the letterRectanlge (no shaddow/ alpha)
      Rectangle innerRectangle = new Rectangle(letterRectangle);
      shrinkRectanlge(texture, innerRectangle, 0.5f);

      character.x = innerRectangle.x - letterRectangle.x;
      character.width = innerRectangle.width;
      character.y = letterRectangle.y - searchPosition.y;

      // update font top and bottom (base line)
      fontTop = Math.min(fontTop, innerRectangle.y - searchPosition.y);
      fontBottom = Math.max(fontBottom, innerRectangle.y + innerRectangle.height - searchPosition.y);

      letter_nr++;
      searchPosition.x = letterRectangle.x + letterRectangle.width;

      characters[character.characterName] = character;
    }
    texture.updateByteBuffer();
  }

  /**
   * Returns the position of the next letter (updates searchPosition).
   *
   * @param texture texture to search in
   * @param searchPosition position to start searching
   * @param lineHeight height of the lines
   * @return position of the next letter or null
   */
  private Point getNextLetterPosition(SmallTexture texture, Point searchPosition, int lineHeight){
    BufferedImage image = texture.getImage();
    int width = image.getWidth();
    int height = image.getHeight();

    while(searchPosition.y < height){
      Rectangle searchRect = new Rectangle(searchPosition.x, searchPosition.y - lineHeight / 4, 1, lineHeight / 2);

      if(!isEmpty(texture, searchRect, 0.0f)) return getNonEmpty(texture, searchRect, 0.0f);

      searchPosition.x++;
      // next line ?
      if(searchPosition.x >= width){
        searchPosition.x = 0;
        searchPosition.y += lineHeight;
      }
    }

    return null;
  }

  /**
   * Extracts the letter from the given position.
   *
   * @param texture texture
   * @param letterPosition position of the letter
   */
  private Rectangle getLetterRectangle(SmallTexture texture, Point letterPosition){
    Rectangle letterRectangle = new Rectangle(letterPosition.x, letterPosition.y, 1, 1);

    growRectanlge(texture, letterRectangle, 0.0f);

    return letterRectangle;
  }

  /**
   * Lets the recangle grow if there are values around with alpha > minAlpha.
   *
   * @param texture texture
   * @param rectangle rectanlge
   */
  private void growRectanlge(SmallTexture texture, Rectangle rectangle, float minAlpha){
    // region growing
    boolean ready = false;
    while(!ready){
      ready = true;

      Rectangle left = new Rectangle(rectangle.x - 1, rectangle.y - 1, 1, rectangle.height + 2);
      Rectangle right = new Rectangle(rectangle.x + rectangle.width, rectangle.y - 1, 1, rectangle.height + 2);
      Rectangle top = new Rectangle(rectangle.x - 1, rectangle.y - 1, rectangle.width + 2, 1);
      Rectangle bottom = new Rectangle(rectangle.x - 1, rectangle.y + rectangle.height, rectangle.width + 2, 1);

      if(!isEmpty(texture, left, minAlpha)){
        rectangle.x--;
        rectangle.width++;
        ready = false;
      }
      if(!isEmpty(texture, right, minAlpha)){
        rectangle.width++;
        ready = false;
      }
      if(!isEmpty(texture, top, minAlpha)){
        rectangle.y--;
        rectangle.height++;
        ready = false;
      }
      if(!isEmpty(texture, bottom, minAlpha)){
        rectangle.height++;
        ready = false;
      }
    }
  }

  /**
   * Lets the recangle shrink if there are values with alpha <= minAlpha.
   *
   * @param texture texture
   * @param rectangle rectanlge
   */
  private void shrinkRectanlge(SmallTexture texture, Rectangle rectangle, float minAlpha){
    boolean ready = false;
    while(!ready){
      Rectangle shrinkedRectangle = new Rectangle(rectangle.x + 1, rectangle.y + 1,
                                                  rectangle.width - 2, rectangle.height - 2);
      growRectanlge(texture, shrinkedRectangle, minAlpha);

      if(shrinkedRectangle.equals(rectangle)) ready = true;

      rectangle.x = shrinkedRectangle.x;
      rectangle.y = shrinkedRectangle.y;
      rectangle.width = shrinkedRectangle.width;
      rectangle.height = shrinkedRectangle.height;
    }
  }

  /**
   * Checks if the given rectangle is empty.
   *
   * @param texture texture
   * @param rectangle rectangle
   * @param minAlpha pixels with alpha > minAlpha are != 0
   * @return true if the rectangle is empty
   */
  private boolean isEmpty(SmallTexture texture, Rectangle rectangle, float minAlpha){
    return getNonEmpty(texture, rectangle, minAlpha) == null;
  }

  /**
   * Returns a point within rectangle with getRGB(point) != 0.
   *
   * @param texture texture
   * @param rectangle rectangle
   * @param minAlpha pixels with alpha > minAlpha are != 0
   * @return null if there is no such point with getRGB(point) != 0
   */
  private Point getNonEmpty(SmallTexture texture, Rectangle rectangle, float minAlpha){
    BufferedImage image = texture.getImage();
    int width = image.getWidth();
    int height = image.getHeight();

    for(int y = rectangle.y; y < rectangle.y + rectangle.height; y++){
      if(y < 0 || y >= height) continue;
      for(int x = rectangle.x; x < rectangle.x + rectangle.width; x++){
        if(x < 0 || x >= width) continue;

        float alpha = ((image.getRGB(x, y) & 0xFF000000l) >> 24) / 255.f;
        if(alpha > minAlpha) return new Point(x, y);
      }
    }

    return null;
  }

  /**
   * Returns the height of the bimap font.
   *
   * @return height of the font
   */
  public int getHeight(){
    return fontBottom - fontTop;
  }

  /**
   * Returns the width of the given string.
   *
   * @param string string
   * @return width
   */
  public int getWidth(String string){
    int stringWidth = 0;

    for(int character_nr = 0; character_nr < string.length(); character_nr++){
      BitmapFontCharacter bitmapCharacter = characters[string.charAt(character_nr)];
      stringWidth += bitmapCharacter != null ? bitmapCharacter.width : spaceWidth;
    }

    return stringWidth;
  }

  /**
   * Draws the string at the given position.
   *
   * @param string string
   * @param x x
   * @param y y
   */
  public void drawString(String string, int x, int y){
    for(int character_nr = 0; character_nr < string.length(); character_nr++){
      BitmapFontCharacter bitmapCharacter = characters[string.charAt(character_nr)];

      if(bitmapCharacter == null){
        x += spaceWidth;
        continue;
      }

      bitmapCharacter.texture.draw(x + bitmapCharacter.x, y + bitmapCharacter.y - fontTop, 1.0f, 1.0f, 0.0f);
      x += bitmapCharacter.width;
    }
  }
}
